/*
 * The Minimal snprintf() implementation
 *
 * Copyright (c) 2013,2014 Michal Ludvig <michal@logix.cz>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the auhor nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ----
 *
 * This is a minimal snprintf() implementation optimised
 * for embedded systems with a very limited program memory.
 * mini_snprintf() doesn't support _all_ the formatting
 * the glibc does but on the other hand is a lot smaller.
 * Here are some numbers from my STM32 project (.bin file size):
 *      no snprintf():      10768 bytes
 *      mini snprintf():    11420 bytes     (+  652 bytes)
 *      glibc snprintf():   34860 bytes     (+24092 bytes)
 * Wasting nearly 24kB of memory just for snprintf() on
 * a chip with 32kB flash is crazy. Use mini_snprintf() instead.
 *
 * 
 * ----
 *
 * Modified 2016 by Michael Harwerth <miq1_AT_gmx_DOT_de>
 *
 * Gratefully took Michal's code, adapted it for Arduinos and added some tiny enhancements:
 *     * basic float format supported ('%5.2f')
 *     * hexadecimal format supported ('%04X')
 *     * ragged left/right supported ('%-2d')
 *     * leading zeroes supported ('%02d')
 *     * hex dump supported ('%h'/'%H' without and '%b'/'%B' with separating blanks
 *       Note: the length with %b/%B/%h/%H formats does not apply to a single byte, but
 *       gives the number of bytes to dump.
 *       '%016B' means: 16 bytes to dump, zero-padded, blanks between, uppercase A, B, C, D, E and F
 *       BE WARNED: mini_snprintf() does not care about memory boundaries!
 *     * mini_uprintf() added to print to any UART
 *     * comment out the #defines NEED_FLOAT, NEED_SERIAL and/or NEED_HEXDUMP to save some more bytes
 *
 */
#include "Arduino.h"

#include <stdarg.h>
#include "mini-printf.h"

#define NEED_FLOAT
#define NEED_HEXDUMP
#define NEED_SERIAL

static char *pbuffer = 0; // global working pointer

static unsigned int
mini_strlen(const char *s)
{
   unsigned int len = 0;
   while (s[len++]);
   return len-1;
}

static unsigned int
mini_itoa(int value, unsigned int radix, unsigned int uppercase, unsigned int unsig,
    char *buffer)
{
   char   *pb = buffer;
   int   negative = 0;
   unsigned int   i, len;

   /* No support for unusual radixes. */
   if (radix > 16)
      return 0;

   if (value < 0 && !unsig) {
      negative = 1;
      value = -value;
   }

   /* This builds the string back to front ... */
   do {
      int digit = value % radix;
      *(pb++) = (digit < 10 ? '0' + digit : (uppercase ? 'A' : 'a') + digit - 10);
      value /= radix;
   } while (value > 0);

   if (negative)
      *(pb++) = '-';

   *(pb) = '\0';

   /* ... now we reverse it (could do it recursively but will
    * conserve the stack space) */
   len = (pb - buffer);
   for (i = 0; i < len / 2; i++) {
      char j = buffer[i];
      buffer[i] = buffer[len-i-1];
      buffer[len-i-1] = j;
   }

   return len;
}

int _putc(char ch, char* buffer, unsigned int buffer_len)
{
   if ((unsigned int)((pbuffer - buffer) + 1) >= buffer_len)
      return 0;
   *(pbuffer++) = ch;
   *(pbuffer) = '\0';
   return 1;
}

int _puts(char *s, unsigned int len, unsigned int runlen, char padchar, char left_pad, char* buffer, unsigned int buffer_len)
{
   unsigned int i;
   unsigned int slen = len;

   if(runlen && (runlen<len || (runlen>len && padchar))) len = runlen;

   if (buffer_len - (pbuffer - buffer) - 1 < len)
      len = buffer_len - (pbuffer - buffer) - 1;
   if(slen>len) slen = len;

   /* Copy to buffer */
   if(left_pad)
   {
      for (i=0;i<len-slen; i++)
         *(pbuffer++) = padchar;
      for (;i<len;i++)
         *(pbuffer++) = s[i-len+slen];
   }
   else
   {
      for (i = 0; i < slen; i++)
         *(pbuffer++) = s[i];
      for (;i < len; i++)
         *(pbuffer++) = padchar;
   } 
   *(pbuffer) = '\0';
   return len;
}

int mini_vsnprintf(char *buffer, unsigned int buffer_len, const char *fmt, va_list va)
{
   char bf[24];
   char ch;
   pbuffer = buffer;

   while ((ch=*(fmt++))) {
      if ((unsigned int)((pbuffer - buffer) + 1) >= buffer_len)
         break;
      if (ch!='%')
         _putc(ch, buffer, buffer_len);
      else {
         char zero_pad = 0;
         char left_pad = 1;
         char *ptr;
         char i = 0;
         char padchar = ' ';
         unsigned int len;
#ifdef NEED_HEXDUMP
         int bytes = 0;
#endif
#ifdef NEED_FLOAT
         char precision = 2;
         int n = 0;
         float f = 0.0;
         float r = 0.0;
#endif

         ch=*(fmt++);

         /* padding requested */
         if (ch=='-')  /* left side padding */
         {
            left_pad = 0;
            ch=*(fmt++);
         }
         if (ch=='0')  /* zero padding */
         {
            padchar = '0';
            ch=*(fmt++);
         }
         if (ch>='0' && ch<='9') {
            do
            {
               zero_pad *=10;
               zero_pad += ch - '0';
               ch=*(fmt++);
            } while(ch>='0' && ch <= '9');
         }
#ifdef NEED_FLOAT
         if (ch=='.')  /* precision */
         {
            ch=*(fmt++);
         }
         if (ch>='0' && ch<='9') {
            precision = ch - '0';
            ch=*(fmt++);
         }
#endif

         switch (ch) {
            case 0:
               goto end;

            case 'u':
            case 'd':
               len = mini_itoa(va_arg(va, int), 10, 0, (ch=='u'), bf);
               _puts(bf, len, zero_pad, padchar, left_pad, buffer, buffer_len);
               break;

            case 'x':
            case 'X':
               len = mini_itoa(va_arg(va, unsigned int), 16, (ch=='X'), 1, bf);
               _puts(bf, len, zero_pad, padchar, left_pad, buffer, buffer_len);
               break;

            case 'c' :
               _putc((char)(va_arg(va, int)), buffer, buffer_len);
               break;

            case 's' :
               ptr = va_arg(va, char*);
               _puts(ptr, mini_strlen(ptr), zero_pad, ' ', left_pad, buffer, buffer_len);
               break;

#ifdef NEED_HEXDUMP
            case 'b':
            case 'B':
            case 'h':
            case 'H':
               ptr = va_arg(va, char*);
	       bytes = zero_pad;
	       while(bytes--)
	       {
                  len = mini_itoa(*ptr, 16, ((ch=='B') || (ch=='H')), 1, bf);
                  i = _puts(bf, len, 2, '0', 1, buffer, buffer_len);
		  if(!i) break;
		  ptr++;
                  if(bytes && ((ch=='b') || (ch=='B'))) _putc(' ', buffer, buffer_len);
	       }
               break;
#endif

#ifdef NEED_FLOAT
            case 'f':
	       r = 0.5;
	       for(i=0;i<precision;++i) r /= 10;
               f = (float)va_arg(va, double);
	       if(f>0) f += r;
	       else    f -= r;
               n = (int)f;
               len = mini_itoa(n, 10, 0, 0, bf);
               _puts(bf, len, zero_pad, padchar, left_pad, buffer, buffer_len);
               if(precision)
               {
                   _putc('.', buffer, buffer_len);
                   f -= n;
		   if(f<0) f = -f;
                   for(i=0;i<precision;++i)
                   {
                      f *= 10.0;
                   }
                   n = (int)f;
                  len = mini_itoa(n, 10, 0, 1, bf);
                  _puts(bf, len, precision, '0', 0, buffer, buffer_len);
               }
               break;
#endif

            default:
               _putc(ch, buffer, buffer_len);
               break;
         }
      }
   }
end:
   return pbuffer - buffer;
}


int
mini_snprintf(char* buffer, unsigned int buffer_len, const char *fmt, ...)
{
   int ret;
   va_list va;
   va_start(va, fmt);
   ret = mini_vsnprintf(buffer, buffer_len, fmt, va);
   va_end(va);

   return ret;
}

#ifdef NEED_SERIAL
int mini_uprintf(Print &serial, const char *fmt, ...)
{
   int ret;
   va_list va;
   char buffer[128];

   va_start(va, fmt);
   ret = mini_vsnprintf(buffer, 128, fmt, va);
   va_end(va);
   serial.print(buffer);

   return ret;
}
#endif
